@@ -39,6 +39,8 @@
 #include <netinet/in.h>
 #ifdef __DragonFly__
 #  include <netproto/802_11/ieee80211_ioctl.h>
+#elif __APPLE__
+  /* FIXME: Add apple includes so we can work out SSID */
 #else
 #  include <net80211/ieee80211_ioctl.h>
 #endif
@@ -59,15 +61,15 @@
 #include "if-options.h"
 #include "net.h"
 
-#define ROUNDUP(a)							      \
+#ifndef RT_ROUNDUP
+#define RT_ROUNDUP(a)							      \
 	((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))
-#define ADVANCE(x, n) (x += ROUNDUP((n)->sa_len))
+#define RT_ADVANCE(x, n) (x += RT_ROUNDUP((n)->sa_len))
+#endif
 
 /* FIXME: Why do we need to check for sa_family 255 */
 #define COPYOUT(sin, sa)						      \
-	sin.s_addr = ((sa) != NULL && ((sa)->sa_family == AF_INET ||	      \
-		(sa)->sa_family == 255))				      \
-	    ?								      \
+	sin.s_addr = ((sa) != NULL) ?					      \
 	    (((struct sockaddr_in *)(void *)sa)->sin_addr).s_addr : 0
 
 static int r_fd = -1;
@@ -198,7 +200,7 @@ if_route(const struct interface *iface, const struct in_addr *dest,
 	int retval = 0;
 
 #define ADDSU(_su) {							      \
-		l = ROUNDUP(_su.sa.sa_len);				      \
+		l = RT_ROUNDUP(_su.sa.sa_len);				      \
 		memcpy(bp, &(_su), l);					      \
 		bp += l;						      \
 	}
@@ -298,7 +300,7 @@ get_addrs(int type, char *cp, struct sockaddr **sa)
 			    inet_ntoa(((struct sockaddr_in *)sa[i])->
 				sin_addr));
 #endif
-			ADVANCE(cp, sa[i]);
+			RT_ADVANCE(cp, sa[i]);
 		} else
 			sa[i] = NULL;
 	}
@@ -317,6 +319,10 @@ manage_link(int fd)
 	struct rt rt;
 	struct sockaddr *sa, *rti_info[RTAX_MAX];
 	int len;
+#ifdef RTM_CHGADDR
+	struct sockaddr_dl sdl;
+	unsigned char *hwaddr;
+#endif
 
 	for (;;) {
 		if (ioctl(fd, FIONREAD, &len) == -1)
@@ -340,6 +346,7 @@ manage_link(int fd)
 		for (p = link_buf; p < e; p += rtm->rtm_msglen) {
 			rtm = (struct rt_msghdr *)(void *)p;
 			switch(rtm->rtm_type) {
+#ifdef RTM_IFANNOUNCE
 			case RTM_IFANNOUNCE:
 				ifan = (struct if_announcemsghdr *)(void *)p;
 				switch(ifan->ifan_what) {
@@ -351,6 +358,7 @@ manage_link(int fd)
 					break;
 				}
 				break;
+#endif
 			case RTM_IFINFO:
 				ifm = (struct if_msghdr *)(void *)p;
 				memset(ifname, 0, sizeof(ifname));
@@ -376,17 +384,41 @@ manage_link(int fd)
 				COPYOUT(rt.gate, rti_info[RTAX_GATEWAY]);
 				route_deleted(&rt);
 				break;
-			case RTM_DELADDR:
+#ifdef RTM_CHGADDR
+			case RTM_CHGADDR:	/* FALLTHROUGH */
+#endif
+			case RTM_DELADDR:	/* FALLTHROUGH */
 			case RTM_NEWADDR:
 				ifam = (struct ifa_msghdr *)(void *)p;
+				if (!if_indextoname(ifam->ifam_index, ifname))
+					break;
 				cp = (char *)(void *)(ifam + 1);
 				get_addrs(ifam->ifam_addrs, cp, rti_info);
-				COPYOUT(rt.dest, rti_info[RTAX_IFA]);
-				COPYOUT(rt.net, rti_info[RTAX_NETMASK]);
-				COPYOUT(rt.gate, rti_info[RTAX_BRD]);
-				if (if_indextoname(ifam->ifam_index, ifname))
+				if (rti_info[RTAX_IFA] == NULL)
+					break;
+				switch (rti_info[RTAX_IFA]->sa_family) {
+#ifdef RTM_CHGADDR
+				case AF_LINK:
+					if (rtm->rtm_type != RTM_CHGADDR)
+						break;
+					memcpy(&sdl, rti_info[RTAX_IFA],
+					    rti_info[RTAX_IFA]->sa_len);
+					hwaddr = xmalloc(sdl.sdl_alen);
+					memcpy(hwaddr, LLADDR(&sdl),
+					    sdl.sdl_alen);
+					handle_hwaddr(ifname, hwaddr,
+					    sdl.sdl_alen);
+					break;
+#endif
+				case AF_INET:
+				case 255: /* FIXME: Why 255? */
+					COPYOUT(rt.dest, rti_info[RTAX_IFA]);
+					COPYOUT(rt.net, rti_info[RTAX_NETMASK]);
+					COPYOUT(rt.gate, rti_info[RTAX_BRD]);
 					handle_ifa(rtm->rtm_type, ifname,
 					    &rt.dest, &rt.net, &rt.gate);
+					break;
+				}
 				break;
 			}
 		}
